#include <string.h>
#include <math.h>
#include "qhmc_qopqdp_common.h"

static char *mtname = "qopqdp.squark";

squark_t *
qopqdp_squark_check(lua_State *L, int idx)
{
  luaL_checkudata(L, idx, mtname);
  squark_t *q = lua_touserdata(L, idx);
  return q;
}

void
qopqdp_squark_array_check(lua_State *L, int idx, int n, squark_t *q[n])
{
  luaL_checktype(L, idx, LUA_TTABLE);
  qassert(lua_objlen(L, idx)==n);
  lua_pushvalue(L, idx); // make copy for indexing convenience
  for(int i=0; i<n; i++) {
    lua_pushnumber(L, i+1);
    lua_gettable(L, -2);
    q[i] = qopqdp_squark_check(L, -1);
    lua_pop(L, 1);
  }
  lua_pop(L, 1);
}

static void
qopqdp_squark_free(lua_State *L, int idx)
{
  squark_t *q = qopqdp_squark_check(L, idx);
  QDP_destroy_V(q->cv);
}

static int
qopqdp_squark_gc(lua_State *L)
{
  qopqdp_squark_free(L, -1);
  return 0;
}

static int
qopqdp_squark_clone(lua_State *L)
{
  BEGIN_ARGS;
  GET_SQUARK(q2);
  OPT_SUBSET(sub, q2->lat, QDP_all_L(q2->qlat));
  END_ARGS;
  squark_t *q1 = qopqdp_squark_create_unset(L, q2->nc, q2->lat);
  QDP_V_eq_V(q1->cv, q2->cv, sub);
  return 1;
}

static int
qopqdp_squark_zero(lua_State *L)
{
  BEGIN_ARGS;
  GET_SQUARK(q);
  OPT_SUBSET(sub, q->lat, QDP_all_L(q->qlat));
  END_ARGS;
  QDP_V_eq_zero(q->cv, sub);
  return 0;
}

// color starts at 1
static int
qopqdp_squark_point(lua_State *L)
{
  BEGIN_ARGS;
  GET_SQUARK(q);
  GET_TABLE_LEN_INDEX(nd,ip);
  GET_INT(color);
  GET_AS_COMPLEX(z);
  END_ARGS;
  int point[nd]; get_int_array(L, ip, nd, point);
  int node = QDP_node_number_L(q->qlat, point);
  if(node==QDP_this_node) {
    int index = QDP_index_L(q->qlat, point);
    QLA_ColorVector *qcv = QDP_site_ptr_readwrite_V(q->cv, index);
    QLA_c_eq_r_plus_ir(QLA_elem_V(*qcv,color-1), z.r, z.i);
  }
  return 0;
}

static int
qopqdp_squark_random(lua_State *L)
{
  BEGIN_ARGS;
  GET_SQUARK(q);
  OPT_SUBSET(sub, q->lat, QDP_all_L(q->qlat));
  END_ARGS;
  QDP_V_eq_gaussian_S(q->cv, qopqdp_srs, sub);
  QLA_Real r = sqrt(0.5); // normalize to sigma^2 = 1/2
  QDP_V_eq_r_times_V(q->cv, &r, q->cv, sub);
  return 0;
}

static void
lnormalize_V(NCPROT QLA_ColorVector(*x), int i)
{
  for(int ic=0; ic<QLA_Nc; ic++) {
    QLA_Complex z = QLA_elem_V(*x,ic);
    QLA_Real n = QLA_norm2_c(z);
    if(n!=0) n = 1/sqrt(n);
    QLA_c_eq_r_times_c(QLA_elem_V(*x,ic), n, z);
  }
}

static int
qopqdp_squark_randomU1(lua_State *L)
{
  BEGIN_ARGS;
  GET_SQUARK(q);
  OPT_SUBSET(sub, q->lat, QDP_all_L(q->qlat));
  END_ARGS;
  QDP_V_eq_gaussian_S(q->cv, qopqdp_srs, sub);
  QDP_V_eq_funci(q->cv, lnormalize_V, sub);
  return 0;
}

static void
lnormalizeZ2_V(NCPROT QLA_ColorVector(*x), int i)
{
  for(int ic=0; ic<QLA_Nc; ic++) {
    QLA_Complex z = QLA_elem_V(*x,ic);
    QLA_Real re = QLA_real(z);
    if (re > 0) { re = 1; } else { re = -1; }
    QLA_c_eq_r_plus_ir(QLA_elem_V(*x,ic), re, 0.0);
  }
}


static int
qopqdp_squark_randomZ2(lua_State *L)
{
  BEGIN_ARGS;
  GET_SQUARK(q);
  OPT_SUBSET(sub, q->lat, QDP_all_L(q->qlat));
  END_ARGS;
  QDP_V_eq_gaussian_S(q->cv, qopqdp_srs, sub);
  QDP_V_eq_funci(q->cv, lnormalizeZ2_V, sub);
  return 0;
}


// 1: squark
// 2: filename
// return file, record metadata
static int
qopqdp_squark_load(lua_State *L)
{
#define NC nc
  BEGIN_ARGS;
  GET_SQUARK(q);
  GET_STRING(fn);
  END_ARGS;
  //printf0("loading lattice file %s\n", fn);
  //double dt = -QDP_time();
  QDP_String *fmd = QDP_string_create();
  QDP_String *rmd = QDP_string_create();
  QDP_Reader *qr = QDP_open_read(fmd, (char*)fn);
  int prec, type, nc;
  qopqdp_get_prec_type_nc(qr, &prec, &type, &nc);
  if(prec=='F') {
#if QDP_Precision == 'F'
    QDP_F_read_V(qr, rmd, q->cv);
#else
    QDP_F_ColorVector *tv;
	tv = QDP_F_create_V();
	QDP_F_read_V(qr, rmd, tv);
	QDP_DF_V_eq_V(q->cv, tv, QDP_all);
	QDP_F_destroy_V(tv);
#endif
  } else { // record precision is 'D'
#if QDP_Precision == 'F'
	QDP_D_ColorVector *tv;
	tv = QDP_D_create_V();
	QDP_D_read_V(qr, rmd, tv);
	QDP_FD_V_eq_V(q->cv, tv, QDP_all);
	QDP_D_destroy_V(tv);
#else
	QDP_D_read_V(qr, rmd, q->cv);
#endif
  }
  QDP_close_read(qr);
  lua_pushstring(L, QDP_string_ptr(fmd));
  lua_pushstring(L, QDP_string_ptr(rmd));
  QDP_string_destroy(fmd);
  QDP_string_destroy(rmd);
  //dt += QDP_time();
  //printf0(" loaded in %g seconds\n", dt);
  return 2;
#undef NC
}

// 1: squark
// 2: filename
// 3: file metadata
// 4: record metadata
// 5: (optional) precision
static int
qopqdp_squark_save(lua_State *L)
{
#define NC nc
  BEGIN_ARGS;
  GET_SQUARK(q);
  GET_STRING(fn);
  GET_STRING(fmds);
  GET_STRING(rmds);
  OPT_STRING(precision, "F");
  END_ARGS;
  //printf0("saving lattice file %s\n", fn);
  //double dt = -QDP_time();
  QDP_String *fmd = QDP_string_create();
  QDP_string_set(fmd, (char *)fmds);
  QDP_String *rmd = QDP_string_create();
  QDP_string_set(rmd, (char *)rmds);
  QDP_Writer *qw = QDP_open_write(fmd, (char*)fn, QDP_SINGLEFILE);
  if(precision[0]=='F') {
#if QDP_Precision == 'F'
    QDP_write_V(qw, rmd, q->cv);
#else
	QDP_F_ColorVector *tv;
	tv = QDP_F_create_V();
	QDP_FD_V_eq_V(tv, q->cv, QDP_all);
	QDP_F_write_V(qw, rmd, tv);
	QDP_F_destroy_V(tv);
#endif
  } else { // record precision is 'D'
#if QDP_Precision == 'F'
	QDP_D_ColorVector *tv;
	tv = QDP_D_create_V();
	QDP_DF_V_eq_V(tv, q->cv, QDP_all);
	QDP_D_write_V(qw, rmd, tv);
	QDP_D_destroy_V(tv);
#else
    QDP_write_V(qw, rmd, q->cv);
#endif
  }
  QDP_close_write(qw);
  QDP_string_destroy(fmd);
  QDP_string_destroy(rmd);
  //dt += QDP_time();
  //printf0(" saved in %g seconds\n", dt);
  return 0;
#undef NC
}

static int
qopqdp_squark_set(lua_State *L)
{
  BEGIN_ARGS;
  GET_SQUARK(q1);
  GET_SQUARK(q2);
  OPT_SUBSET(sub, q1->lat, QDP_all_L(q1->qlat));
  END_ARGS;
  QDP_V_eq_V(q1->cv, q2->cv, sub);
  return 0;
}

static int
qopqdp_squark_conj(lua_State *L)
{
  BEGIN_ARGS;
  GET_SQUARK(q1);
  GET_SQUARK(q2);
  OPT_SUBSET(sub, q1->lat, QDP_all_L(q1->qlat));
  END_ARGS;
  QDP_V_eq_conj_V(q1->cv, q2->cv, sub);
  return 0;
}

struct squark_intwall_info
{
  int val; // Should always be 1.
  int timeslice;
  int td;
};

// Value contains a timeslice and the time direction.
void
private_squark_intwall(QLA_Int *i, int coords[], void *value)
{
  struct squark_intwall_info data = *(struct squark_intwall_info*)value;
  if (coords[data.td] == data.timeslice) {
    QLA_I_eq_I(i, &(data.val));
  }
}

// Create a function that handles the dilution.
void
private_squark_dilute(QDP_ColorVector* cv_dil, QDP_ColorVector* cv_input, int timeslice, int color, int space_eo)
{
  QDP_ColorVector* temp_cvec = QDP_create_V();
  QDP_V_eq_zero(temp_cvec, QDP_all);
  
  // First, get a subset right.
  QDP_Subset subset_use = QDP_all;
  if (space_eo == 0)
  {
	subset_use = QDP_even;
  }
  else if (space_eo == 1)
  {
    subset_use = QDP_odd;
  }
  
  // Next, process color (and even/odd, technically).
  if (color != -1)
  {
    QDP_Complex* temp_cplx = QDP_create_C();
	QDP_C_eq_zero(temp_cplx, QDP_all);
	QDP_C_eq_elem_V(temp_cplx, cv_input, color, subset_use);
	QDP_V_eq_elem_C(temp_cvec, temp_cplx, color, subset_use);
    QDP_destroy_C(temp_cplx);
  }
  else
  {
    QDP_V_eq_V(temp_cvec, cv_input, subset_use);
  }
  
  // Last, process timeslice dilution.
  if (timeslice == -1) // no timeslice dilution.
  {
	QDP_V_eq_V(cv_dil, temp_cvec, subset_use);
  }
  else // Create an integer wall.
  {
    QDP_Int* latint = QDP_create_I();
	QDP_I_eq_zero(latint, QDP_all);
	
	struct squark_intwall_info data;
	data.timeslice = timeslice;
    data.td = 3;
	data.val = 1;
	
	// Build a mask that has 1's on the timeslice you want, 0's elsewhere.
	QDP_I_eq_funca(latint, private_squark_intwall, (void*)(&data), QDP_all);
	
	// And now copy.
	QDP_V_eq_V_mask_I(cv_dil, temp_cvec, latint, subset_use);
	
	// And clean up.
	QDP_destroy_I(latint);
  }
  
  QDP_destroy_V(temp_cvec);

}

// ESW addition 03/20/2014
// Pass in a quark (arg1), and a set of values (arg2, arg3, arg4)
// -> (time, color, space even/odd). -1 means don't dilute. Otherwise,
// only copy values with that timeslice/color/even or odd.
// ESW fix 05/14/2014 -- close a memory leak.
static int
qopqdp_squark_dilute(lua_State *L)
{
  BEGIN_ARGS;
  GET_SQUARK(q1); // diluted vector goes here.
  GET_SQUARK(q2);
  GET_INT(timeslice); // a time coordinate, 0 normalized.
  GET_INT(color); // a color coordinate, 0 normalized.
  GET_INT(space_eo); // even-odd -> 0-1
  END_ARGS;
  
  //squark_t *q1 = qopqdp_squark_create_unset(L, q2->nc, q2->lat);
  QDP_V_eq_zero(q1->cv, QDP_all);
  
  // Call an encapsulating function.
  private_squark_dilute(q1->cv, q2->cv, timeslice, color, space_eo);
  
  return 0;
}

static int
qopqdp_squark_combine(lua_State *L)
{
  int narg = lua_gettop(L);
  qassert(narg==3 || narg==4);
  squark_t *qd = qopqdp_squark_check(L, 1);
  int nqs; get_table_len(L, 2, &nqs);
  squark_t *qs[nqs]; qopqdp_squark_array_check(L, 2, nqs, qs);
  int nc; get_table_len(L, 3, &nc);
  qassert(nqs==nc);
  QDP_Subset sub = QDP_all;
  if(narg>3) {
    sub = qopqdp_check_subset(L, 4, qd->lat);
  }
  for(int i=0; i<nqs; i++) {
    lua_pushinteger(L, i+1);
    lua_gettable(L, 3);
    if(lua_type(L,-1)==LUA_TNUMBER) {
      QLA_Real r = lua_tonumber(L, -1);
      if(i==0) QDP_V_eq_r_times_V(qd->cv, &r, qs[0]->cv, sub);
      else QDP_V_peq_r_times_V(qd->cv, &r, qs[i]->cv, sub);
    } else { // complex
      qhmc_complex_t *c = qhmc_complex_check(L, -1);
      QLA_Complex z;
      QLA_c_eq_r_plus_ir(z, c->r, c->i);
      if(i==0) QDP_V_eq_c_times_V(qd->cv, &z, qs[0]->cv, sub);
      else QDP_V_peq_c_times_V(qd->cv, &z, qs[i]->cv, sub);
    }
  }
  return 0;
}

static int
qopqdp_squark_getSite(lua_State *L)
{
#define NC QDP_get_nc(q->cv)
  BEGIN_ARGS;
  GET_SQUARK(q);
  GET_TABLE_LEN_INDEX(nd,ip);
  END_ARGS;
  int site[nd]; get_int_array(L, ip, nd, site);
  QLA_ColorVector qcv;
  int node = QDP_node_number_L(q->qlat, site);
  if(node==QDP_this_node) {
    int index = QDP_index_L(q->qlat, site);
    QLA_ColorVector *qcvp = QDP_site_ptr_readonly_V(q->cv, index);
    QLA_V_eq_V(&qcv, qcvp);
  } else {
    QLA_V_eq_zero(&qcv);
  }
  QMP_sum_double_array((double *)&qcv, 2*QLA_Nc);
  push_complex_array(L, QLA_Nc, (qhmc_complex_t *)&qcv);
  return 1;
#undef NC
}

static int
qopqdp_squark_norm2(lua_State *L)
{
  BEGIN_ARGS;
  GET_SQUARK(q);
  OPT_SUBSETS(subs, ns, q->lat, QDP_all_and_empty_L(q->qlat), 1);
  END_ARGS;
  if(ns==1) {
    QLA_Real nrm2;
    QDP_r_eq_norm2_V(&nrm2, q->cv, subs[0]);
    lua_pushnumber(L, nrm2);
  } else {
    QLA_Real nrm2[ns];
    QDP_r_eq_norm2_V_multi(nrm2, q->cv, subs, ns);
    push_double_array(L, ns, nrm2);
  }
  return 1;
}

static int
qopqdp_squark_norm2_wallsink(lua_State *L)
{
  BEGIN_ARGS;
  GET_SQUARK(q);
  OPT_SUBSETS(subs, ns, q->lat, QDP_all_and_empty_L(q->qlat), 1);
  END_ARGS;
  int i;
  if(ns==1) { // I mean, this doesn't really make sense, but whatever!
    QLA_ColorVector thesum;
	QDP_v_eq_sum_V(&thesum, q->cv, subs[0]);
    QLA_Real nrm2;
	QLA_R_eq_norm2_V(&nrm2, &thesum);
    lua_pushnumber(L, nrm2);
  } else { // This is a wall sink contraction.
    QLA_ColorVector thesums[ns];
	QDP_v_eq_sum_V_multi(thesums, q->cv, subs, ns);
    QLA_Real nrm2[ns];
	for (i=0;i<ns;i++)
	{
	   QLA_R_eq_norm2_V(&(nrm2[i]),&(thesums[i]));
	}
    push_double_array(L, ns, nrm2);
  }
  return 1;
}

static int
qopqdp_squark_redot(lua_State *L)
{
  BEGIN_ARGS;
  GET_SQUARK(q1);
  GET_SQUARK(q2);
  OPT_SUBSETS(subs, ns, q1->lat, QDP_all_and_empty_L(q1->qlat), 1);
  END_ARGS;
  if(ns==1) {
    QLA_Real redot;
    QDP_r_eq_re_V_dot_V(&redot, q1->cv, q2->cv, subs[0]);
    lua_pushnumber(L, redot);
  } else {
    QLA_Real redot[ns];
    QDP_r_eq_re_V_dot_V_multi(redot, q1->cv, q2->cv, subs, ns);
    push_double_array(L, ns, redot);
  }
  return 1;
}

// 1: squark
// 2: squark
// 3: (optional) subset
static int
qopqdp_squark_dot(lua_State *L)
{
  BEGIN_ARGS;
  GET_SQUARK(q1);
  GET_SQUARK(q2);
  OPT_SUBSETS(subs, ns, q1->lat, QDP_all_and_empty_L(q1->qlat), 1);
  END_ARGS;
  if(ns==1) {
    QLA_Complex dot;
    QDP_c_eq_V_dot_V(&dot, q1->cv, q2->cv, subs[0]);
    qhmc_complex_create(L, QLA_real(dot), QLA_imag(dot));
  } else {
    QLA_Complex dot[ns];
    QDP_c_eq_V_dot_V_multi(dot, q1->cv, q2->cv, subs, ns);
    qhmc_complex_t qdot[ns];
    for(int i=0; i<ns; i++) {
      qdot[i].r = QLA_real(dot[i]);
      qdot[i].i = QLA_imag(dot[i]);
    }
    push_complex_array(L, ns, qdot);
  }
  return 1;
}

struct squark_wall_info
{
  QLA_ColorVector value;
  int timeslice;
  int td;
};

// Value must be a pointer to a QLA_ColorVector
void
private_squark_wall(QLA_ColorVector *s, int coords[], void *value)
{
  struct squark_wall_info data = *(struct squark_wall_info*)value;
  if (coords[data.td] == data.timeslice) {
    QLA_V_eq_V(s, &(data.value));
  }
}

// Constructs a wall source. This matches the convention in MILC.
// Arguments:
// 1 - the staggered quark, as always.
// 2 - Time slice.
// 3 - An integer. 0 = even source, 1 = odd source.
// THIS IS DIFFERENT FROM THE DEFINITIONS IN GUPTA ET AL (1991).
// This is even/odd in terms of lattice sites. 
// 4 - Color
// 5 - value (real or complex)
static int
qopqdp_squark_wall(lua_State *L)
{
  BEGIN_ARGS;
  GET_SQUARK(q);
  GET_INT(timeslice);
  GET_INT(evenodd);
  GET_INT(color);
  GET_AS_COMPLEX(z);
  END_ARGS;
  struct squark_wall_info data;
  QLA_V_eq_zero(&data.value);
  QLA_c_eq_r_plus_ir(QLA_elem_V(data.value,color-1), z.r, z.i);
  data.timeslice = timeslice;
  data.td = q->lat->nd-1;
  QDP_V_eq_funca(q->cv, private_squark_wall, (void*)(&data), (evenodd == 0 ? QDP_even : QDP_odd));
  return 0;
}

struct squark_wall_corner_info
{
  QLA_ColorVector value;
  int timeslice;
  int td;
  int delta;
};

// Value must be a pointer to a QLA_ColorVector
void
private_squark_wall_corner(QLA_ColorVector *s, int coords[], void *value)
{
  struct squark_wall_corner_info data = *(struct squark_wall_corner_info*)value;
  if ((4*(coords[2]%2)+2*(coords[1]%2)+coords[0]%2) == data.delta && coords[data.td] == data.timeslice) {
    QLA_V_eq_V(s, &(data.value));
  }
  else
  {
    QLA_V_eq_zero(s);
  }
}

// Constructs a wall (corner of hypercube) source.
// Arguments:
// 1 - the staggered quark, as always.
// 2 - Time slice.
// 3 - An integer for what corner. delta = (nz%2)(ny%2)(nx%2)
// THIS IS DIFFERENT FROM THE DEFINITIONS IN GUPTA ET AL (1991).
// This is even/odd in terms of lattice sites. 
// 4 - Color
// 5 - value (real or complex)
static int
qopqdp_squark_wall_corner(lua_State *L)
{
  BEGIN_ARGS;
  GET_SQUARK(q);
  GET_INT(timeslice);
  GET_INT(delta);
  GET_INT(color);
  GET_AS_COMPLEX(z);
  END_ARGS;
  struct squark_wall_corner_info data;
  QLA_V_eq_zero(&data.value);
  QLA_c_eq_r_plus_ir(QLA_elem_V(data.value,color-1), z.r, z.i);
  data.timeslice = timeslice;
  data.td = q->lat->nd-1;
  data.delta = delta;
  QDP_V_eq_funca(q->cv, private_squark_wall_corner, (void*)(&data), QDP_all);
  return 0;
}

// Value must be a pointer to a QLA_ColorVector
void
private_squark_wall_gaussian(QLA_ColorVector *s, int coords[], void *value)
{
  struct squark_wall_info data = *(struct squark_wall_info*)value;
  if (coords[data.td] == data.timeslice) { // normalize it
    QLA_ColorVector vec;
	QLA_Real norm_val;
	QLA_V_eq_V(&vec, s);
	QLA_R_eq_norm2_V(&norm_val, &vec);
	if (norm_val != 0.0)
	{
	   norm_val = 1.0/sqrt(norm_val);
	   QLA_V_eq_r_times_V(s, &norm_val, &vec);
	}
  }
  else // set it to zero.
  {
    QLA_V_eq_zero(s);
  }
}

// Constructs a normalized gaussian wall source. 
// Arguments:
// 1 - the staggered quark, as always.
// 2 - Time slice.
static int
qopqdp_squark_wall_gaussian(lua_State *L)
{
  BEGIN_ARGS;
  GET_SQUARK(q);
  GET_INT(timeslice);
  END_ARGS;
  struct squark_wall_info data;
  QDP_V_eq_gaussian_S(q->cv, qopqdp_srs, QDP_all);
  QLA_V_eq_zero(&data.value);
  QLA_c_eq_r_plus_ir(QLA_elem_V(data.value,0), 0.0, 0.0); // doesn't matter since it gets a random fill.
  data.timeslice = timeslice;
  data.td = q->lat->nd-1;
  QDP_V_eq_funca(q->cv, private_squark_wall_gaussian, (void*)(&data), QDP_all);
  return 0;
}



/* ESW additions for staggered quark rephasing */

// Define a global parameter to store a phase.
// This being global is sadly unavoidable.
// It's how the gauge field rephasing is also done,
// so it's presumably a safe method to use.
int phase_squark_esw;

// Define a global parameter to store a relative location.
int *relative_loc_esw;

// Here's a function used to modify the source field with vectors.
// Based on void point_V in qdp-1.9.1/examples/sample_ks.c
#define NC nc
static void
private_squark_rephase(NCPROT QLA_ColorVector *s, int coords[])
{
  // Build up the phase.
  int phase = 0;
  int bits = phase_squark_esw;
  for (int i = 0; bits; i++)
    {
      // if the coordinate minus the relative coordinate is odd,
      // and if the phase flag is set, add 1.
      // if the coordinate minus the relative coordinate is odd,
      // and if the phase flag is not set, add 0.
      // if the coordinate minus the relative coordinate is even,
      // always add 0.

      if(bits&1) {
	phase += coords[i] - relative_loc_esw[i];
      }
      bits >>= 1;
    }

  if ((phase%2) == 1) // if the overall phase exponent is odd
    {
      QLA_V_eqm_V(s, s);
    }
}
#undef NC

static int
qopqdp_squark_rephase(lua_State *L)
{
  int narg = lua_gettop(L); // Get the number of arguments.
  qassert(narg==3);
  squark_t *qd = qopqdp_squark_check(L, 1); // Check for a staggered quark.
  int phase_flag = luaL_checkint(L, 2); // Check for the phase flag.
  int nd; get_table_len(L, 3, &nd); // Get the length of the relative coordinate.
  int r0[nd];
  get_int_array(L, 3, nd, r0);
  relative_loc_esw = r0;

  // The phase flag is defined bit-wise.
  // 0 = no phase, 1 = phase
  // (x4_sign -- time) (x3_sign) (x2_sign) (x1_sign)
  phase_squark_esw = phase_flag;

  // Apply the phase.
  QDP_V_eq_func(qd->cv, private_squark_rephase, QDP_all);

  return 0;
}

static int
qopqdp_squark_symshift(lua_State *L)
{
#define NC QDP_get_nc(qs->cv)
  int narg = lua_gettop(L);
  qassert(narg==4);
  squark_t *qd = qopqdp_squark_check(L, 1);
  squark_t *qs = qopqdp_squark_check(L, 2);
  gauge_t *g = qopqdp_gauge_check(L, 3);
  int mu = luaL_checkint(L, 4) - 1;
  QDP_ColorVector *tf = QDP_create_V();
  QDP_ColorVector *tb1 = QDP_create_V();
  QDP_ColorVector *tb2 = QDP_create_V();
  QDP_V_eq_sV(tf, qs->cv, QDP_neighbor[mu], QDP_forward, QDP_all);
  QDP_V_eq_Ma_times_V(tb1, g->links[mu], qs->cv, QDP_all);
  QDP_V_eq_sV(tb2, tb1, QDP_neighbor[mu], QDP_backward, QDP_all);
  QDP_V_eq_M_times_V(qd->cv, g->links[mu], tf, QDP_all);
  QDP_V_peq_V(qd->cv, tb2, QDP_all);
  QDP_destroy_V(tf);
  QDP_destroy_V(tb1);
  QDP_destroy_V(tb2);
  return 0;
#undef NC
}

// ESW addition for computing baryon correlators.
// 1: first quark
// 2: array of Nc-1 quarks
// 3: (optional) sublattice
static int
qopqdp_squark_epscontr(lua_State *L)
{
#define NC nc
  BEGIN_ARGS;
  GET_SQUARK(q1);
  GET_TABLE_LEN_INDEX(nc1,iq);
  int nc = nc1 + 1;
  squark_t *qs[nc]; qs[0]=q1; qopqdp_squark_array_check(L, iq, nc1, qs+1);
  OPT_SUBSETS(subs, ns, qs[0]->lat, QDP_all_and_empty_L(qs[0]->qlat), 1);
  END_ARGS;
  QDP_Subset sub = subs[0];
  if(ns>1) sub = QDP_all_L(qs[0]->qlat);
  // Create a color matrix, insert fermions into it.
  QDP_ColorMatrix *mat = QDP_create_M_L(qs[0]->qlat);
  QDP_Complex *determ = QDP_create_C_L(qs[0]->qlat);
  for(int i=0; i<nc; i++) {
    QDP_M_eq_colorvec_V(mat, qs[i]->cv, i, sub);
  }
  QDP_C_eq_det_M(determ, mat, sub);
  if(ns==1) {
    QLA_Complex val;
    QDP_c_eq_sum_C(&val, determ, sub);
    qhmc_complex_create(L, QLA_real(val), QLA_imag(val));
  } else {
    QLA_Complex val[ns];
    QDP_c_eq_sum_C_multi(val, determ, subs, ns);
    qhmc_complex_t qval[ns];
    for(int i=0; i<ns; i++) {
      qval[i].r = QLA_real(val[i]);
      qval[i].i = QLA_imag(val[i]);
    }
    push_complex_array(L, ns, qval);
  }
  QDP_destroy_M(mat);
  QDP_destroy_C(determ);
  return 1;
}

// used for computing the sigma operators.
static int
qopqdp_squark_sum_lattice(lua_State *L)
{
  BEGIN_ARGS;
  GET_SQUARK(q1);
  GET_INT(color);
  OPT_SUBSETS(subs, ns, q1->lat, QDP_all_and_empty_L(q1->qlat), 1);
  END_ARGS;
  QDP_Complex *components = QDP_create_C_L(q1->qlat);
  if(ns==1) {
    QLA_Complex val;
    QDP_C_eq_elem_V(components, q1->cv, color-1, subs[0]);
	QDP_c_eq_sum_C(&val, components, subs[0]);
    qhmc_complex_create(L, QLA_real(val), QLA_imag(val));
  } else {
    QLA_Complex val[ns];
	QDP_C_eq_elem_V(components, q1->cv, color-1, QDP_all);
    QDP_c_eq_sum_C_multi(val, components, subs, ns);
    qhmc_complex_t qval[ns];
    for(int i=0; i<ns; i++) {
      qval[i].r = QLA_real(val[i]);
      qval[i].i = QLA_imag(val[i]);
    }
    push_complex_array(L, ns, qval);
  }
  QDP_destroy_C(components);
  return 1;
}

static int
qopqdp_squark_s4_ferm_observables(lua_State *L)
{
  int narg = lua_gettop(L);
  qassert(narg==3);
  squark_t *q = qopqdp_squark_check(L, 1);
  squark_t *q2 = qopqdp_squark_check(L, 2);
  gauge_t *g = qopqdp_gauge_check(L, 3);
  int i;
  int r0[4]; // for phases.
  
  QDP_Subset* dirsubset;
  
  QLA_Complex interm[2];
  qhmc_complex_t pbp_e[6], pbp_o[6];
  // 0 - pbp on even and odd sites.
  // 1 - add t one-link pbp if x even/odd
  // 2 - add t one-link pbp if t even/odd
  // 3 - add x one-link pbp if x even/odd
  // 4 - add y one-link pbp if y even/odd
  // 5 - add z one-link pbp if z even/odd
  
  // Because we need it.
  QDP_ColorVector* temp_vec;
  QDP_Complex* temp_cmplx;
  
  // Zero out arrays.
  for (i = 0; i < 6; i++)
  {
    pbp_e[i].r = pbp_e[i].i = pbp_o[i].r = pbp_o[i].i = 0;
  }
  
  // Get even, odd pbp.
  QDP_c_eq_V_dot_V(&interm[0], q->cv, q->cv, QDP_even);
  QDP_c_eq_V_dot_V(&interm[1], q->cv, q->cv, QDP_odd);
  pbp_e[0].r = QLA_real(interm[0]);
  pbp_e[0].i = QLA_imag(interm[0]);
  pbp_o[0].r = QLA_real(interm[1]);
  pbp_o[0].i = QLA_imag(interm[1]);
  
  // Generate one-link separated.
  temp_vec = QDP_create_V();
  temp_cmplx = QDP_create_C();
  
  // Time one link.
  QDP_V_eq_M_times_sV(temp_vec, g->links[3], q->cv, QDP_neighbor[3], QDP_forward, QDP_all);
  QDP_C_eq_V_dot_V(temp_cmplx, q2->cv, temp_vec, QDP_all);
  
  dirsubset = qhmcqdp_get_eodir(q->lat, 0); // x even/odd
  QDP_c_eq_sum_C_multi(interm, temp_cmplx, dirsubset, 2);
  pbp_e[1].r = QLA_real(interm[0]);
  pbp_e[1].i = QLA_imag(interm[0]);
  pbp_o[1].r = QLA_real(interm[1]);
  pbp_o[1].i = QLA_imag(interm[1]);
  
  dirsubset = qhmcqdp_get_eodir(q->lat, 3); // t even/odd
  QDP_c_eq_sum_C_multi(interm, temp_cmplx, dirsubset, 2);
  pbp_e[2].r = QLA_real(interm[0]);
  pbp_e[2].i = QLA_imag(interm[0]);
  pbp_o[2].r = QLA_real(interm[1]);
  pbp_o[2].i = QLA_imag(interm[1]);
  
  // Prepare for phases.
  r0[0] = r0[1] = r0[2] = r0[3] = 0;
  relative_loc_esw = r0;
  
  // X one link. Don't forget phase factors!
  QDP_V_eq_M_times_sV(temp_vec, g->links[0], q->cv, QDP_neighbor[0], QDP_forward, QDP_all);
  phase_squark_esw = 0x8; // time
  QDP_V_eq_func(temp_vec, private_squark_rephase, QDP_all);
  QDP_C_eq_V_dot_V(temp_cmplx, q2->cv, temp_vec, QDP_all);
  dirsubset = qhmcqdp_get_eodir(q->lat, 0); // x even/odd
  QDP_c_eq_sum_C_multi(interm, temp_cmplx, dirsubset, 2);
  pbp_e[3].r = QLA_real(interm[0]);
  pbp_e[3].i = QLA_imag(interm[0]);
  pbp_o[3].r = QLA_real(interm[1]);
  pbp_o[3].i = QLA_imag(interm[1]);
  
  // Y one link.
  QDP_V_eq_M_times_sV(temp_vec, g->links[1], q->cv, QDP_neighbor[1], QDP_forward, QDP_all);
  phase_squark_esw = 0x9; // time, x
  QDP_V_eq_func(temp_vec, private_squark_rephase, QDP_all);
  QDP_C_eq_V_dot_V(temp_cmplx, q2->cv, temp_vec, QDP_all);
  dirsubset = qhmcqdp_get_eodir(q->lat, 1); // y even/odd
  QDP_c_eq_sum_C_multi(interm, temp_cmplx, dirsubset, 2);
  pbp_e[4].r = QLA_real(interm[0]);
  pbp_e[4].i = QLA_imag(interm[0]);
  pbp_o[4].r = QLA_real(interm[1]);
  pbp_o[4].i = QLA_imag(interm[1]);
  
  // Z one link.
  QDP_V_eq_M_times_sV(temp_vec, g->links[2], q->cv, QDP_neighbor[2], QDP_forward, QDP_all);
  phase_squark_esw = 0xB; // time, x, y
  QDP_V_eq_func(temp_vec, private_squark_rephase, QDP_all);
  QDP_C_eq_V_dot_V(temp_cmplx, q2->cv, temp_vec, QDP_all);
  dirsubset = qhmcqdp_get_eodir(q->lat, 2); // z even/odd
  QDP_c_eq_sum_C_multi(interm, temp_cmplx, dirsubset, 2);
  pbp_e[5].r = QLA_real(interm[0]);
  pbp_e[5].i = QLA_imag(interm[0]);
  pbp_o[5].r = QLA_real(interm[1]);
  pbp_o[5].i = QLA_imag(interm[1]);
  
  QDP_destroy_V(temp_vec);
  QDP_destroy_C(temp_cmplx);
  
  push_complex_array(L, 6, pbp_e);
  push_complex_array(L, 6, pbp_o);
  
  return 2;
}

// ESW addition 5/14/2014
// Figure out a dilution index.


// ESW addition 5/13/2014
// See if making the contractions C-level really speeds things up!
static int
qopqdp_squark_stoch_connected(lua_State *L)
{
  // Get everything.
  int narg = lua_gettop(L);
  qassert(narg==6);
  squark_t *qd = qopqdp_squark_check(L, 1);
  int n_eta; get_table_len(L, 2, &n_eta);
  squark_t *eta[n_eta]; qopqdp_squark_array_check(L, 2, n_eta, eta);
  int n_phi; get_table_len(L, 3, &n_phi);
  squark_t *phi[n_phi]; qopqdp_squark_array_check(L, 3, n_phi, phi);
  int an_int = 4;
  int timeflag = qhmc_opt_int(L, &an_int, 1, 0);
  int colorflag = qhmc_opt_int(L, &an_int, 1, 0);
  int spaceflag = qhmc_opt_int(L, &an_int, 1, 0);
  
  
  QDP_Subset* sub = qhmcqdp_get_timeslices(qd->lat);
  
  
  // Learn some things about how much memory we need to allocate.
  int numtime = 1, numcolor = 1, numspace = 1;
  if (timeflag != 0)
  {
	numtime = timeflag;
  }
  if (colorflag != 0)
  {
    numcolor = colorflag;
  }
  if (spaceflag != 0)
  {
    numspace = spaceflag;
  }
  
  int totaldil = numtime*numcolor*numspace; // Total number of dilutions.
  
  // Sanity check.
  qassert((n_eta*totaldil)==n_phi); 
  
  // Local variable inside loop for which source, timeslice, color, and e/o to be on.
  int srcsite, timesite, colorsite, spacesite;
  
  
  // Initialize some variables!
  // We need to get all of the 'operators'.
  // If there are memory issues, they should hit here.
  int nt = QDP_coord_size(qd->lat->nd-1);
  //QLA_Complex ops_pion_gs[n_phi][n_phi][nt];
  QLA_Complex ops_scalar[n_phi][n_phi][nt];
  //QLA_Complex ops_rho_x[n_phi][n_phi][nt];
  //QLA_Complex ops_rho_y[n_phi][n_phi][nt];
  //QLA_Complex ops_rho_z[n_phi][n_phi][nt];
  //QLA_Complex ops_rho_x_g4[n_phi][n_phi][nt];
  //QLA_Complex ops_rho_y_g4[n_phi][n_phi][nt];
  //QLA_Complex ops_rho_z_g4[n_phi][n_phi][nt];
  
  QDP_ColorVector* tmp = QDP_create_V();
  
  // Set the relative term for the phase.
  int location_ph[4] = {0, 0, 0, 0};
  relative_loc_esw = (int*)location_ph;

  // Loop over all the things!
  for (int i = 0; i < n_eta*totaldil; i++)
  {
    int tmpnum = i;
    // Figure out which dilution we need.
	
	// Space dilution.
	if (numspace == 1)
	{
	   spacesite = -1;
	}
	else
	{
	   spacesite = tmpnum%numspace;
	   tmpnum = (tmpnum-spacesite)/numspace;
	}
	
	// Color dilution.
	if (numcolor == 1)
	{
	  colorsite = -1;
	}
	else
	{
	   colorsite = tmpnum%numcolor;
	   tmpnum = (tmpnum-colorsite)/numcolor;
	}
	
	// Timeslice even/odd dilution.
	if (numtime == 1)
	{
	  timesite = -1;
	}
	else
	{
	   timesite = tmpnum%numtime;
	   tmpnum = (tmpnum-timesite)/numtime;
	}
	
	// And last get which source we care about.
	srcsite = tmpnum % n_eta;
	tmpnum = (tmpnum - srcsite)/n_eta;
	
	printf0("Src %i Space %i Color %i Time %i\n", srcsite, spacesite, colorsite, timesite);
	
	qassert(tmpnum==0);
	
	// Get a diluted vector!
	QDP_V_eq_zero(tmp, QDP_all);
	private_squark_dilute(tmp, eta[srcsite]->cv, timesite, colorsite, spacesite);
	   
	for (int j = 0; j < n_phi; j++)
	{
	  // Scalar
	  QDP_c_eq_V_dot_V_multi(ops_scalar[i][j], tmp, phi[j]->cv, sub, nt);
	  /*
	  // Pion GB (-1)^(x+y+z+t)
	  phase_squark_esw = 15; // 0b1111
	  QDP_V_eq_func(tmp, private_squark_rephase, QDP_all);
	  QDP_c_eq_V_dot_V_multi(ops_pion_gs[i][j], tmp, phi[j]->cv, sub, nt);
	  
	  // Rho_x (-1)^(y+z+t)
	  phase_squark_esw = 1; // 0b0001
	  QDP_V_eq_func(tmp, private_squark_rephase, QDP_all);
	  QDP_c_eq_V_dot_V_multi(ops_rho_x[i][j], tmp, phi[j]->cv, sub, nt);
	  
	  // Rho_y (-1)^(x+z+t)
	  phase_squark_esw = 3; // 0b0011
	  QDP_V_eq_func(tmp, private_squark_rephase, QDP_all);
	  QDP_c_eq_V_dot_V_multi(ops_rho_y[i][j], tmp, phi[j]->cv, sub, nt);
	  
	  // Rho_z (-1)^(x+y+t)
	  phase_squark_esw = 6; // 0b0110
	  QDP_V_eq_func(tmp, private_squark_rephase, QDP_all);
	  QDP_c_eq_V_dot_V_multi(ops_rho_z[i][j], tmp, phi[j]->cv, sub, nt);
	  
	  // g4 Rho_x (-1)^(x)
	  phase_squark_esw = 10; // 0b1010
	  QDP_V_eq_func(tmp, private_squark_rephase, QDP_all);
	  QDP_c_eq_V_dot_V_multi(ops_rho_x_g4[i][j], tmp, phi[j]->cv, sub, nt);
	  
	  // g4 Rho_y (-1)^(y)
	  phase_squark_esw = 3; // 0b0011
	  QDP_V_eq_func(tmp, private_squark_rephase, QDP_all);
	  QDP_c_eq_V_dot_V_multi(ops_rho_y_g4[i][j], tmp, phi[j]->cv, sub, nt);
	  
	  // g4 Rho_z (-1)^(z)
	  phase_squark_esw = 6; // 0b0110
	  QDP_V_eq_func(tmp, private_squark_rephase, QDP_all);
	  QDP_c_eq_V_dot_V_multi(ops_rho_z_g4[i][j], tmp, phi[j]->cv, sub, nt);
	  
	  
	  // And reset.
	  phase_squark_esw = 4; // 0b0100
	  QDP_V_eq_func(tmp, private_squark_rephase, QDP_all);*/
	}
  }
  
  QDP_destroy_V(tmp);
  
  // Form all contractions.
  //QLA_Complex corr_pion_gs[nt];
  QLA_Complex corr_scalar[nt];
  //QLA_Complex corr_rho[nt];
  //QLA_Complex corr_rho_g4[nt];
  for (int i=0; i < nt; i++)
  {
	//QLA_C_eq_zero(&corr_pion_gs[i]);
	QLA_C_eq_zero(&corr_scalar[i]);
	//QLA_C_eq_zero(&corr_rho[i]);
	//QLA_C_eq_zero(&corr_rho_g4[i]);
  }
  
  for (int i1=0; i1 < n_eta-1; i1++)
  {
	for (int j1=i1+1; j1<n_eta; j1++)
	{
	  for (int i2=0; i2< totaldil; i2++)
	  {
	    for (int j2=0; j2 < totaldil; j2++)
		{
		  int i = i1*totaldil+i2;
		  int j = j1*totaldil+j2;
	        for (int t=0; t<nt; t++)
	        {
	          for (int t0 = 0; t0 < nt; t0++)
		      {
      		    //QLA_C_peq_C_times_C(&corr_pion_gs[t0], &ops_pion_gs[i][j][t], &ops_pion_gs[j][i][(t+t0)%nt]);
		        QLA_C_peq_C_times_C(&corr_scalar[t0], &ops_scalar[i][j][t], &ops_scalar[j][i][(t+t0)%nt]);
		        /*QLA_C_peq_C_times_C(&corr_rho[t0], &ops_rho_x[i][j][t], &ops_rho_x[j][i][(t+t0)%nt]);
		        QLA_C_peq_C_times_C(&corr_rho[t0], &ops_rho_y[i][j][t], &ops_rho_y[j][i][(t+t0)%nt]);
		        QLA_C_peq_C_times_C(&corr_rho[t0], &ops_rho_z[i][j][t], &ops_rho_z[j][i][(t+t0)%nt]);
		        QLA_C_peq_C_times_C(&corr_rho_g4[t0], &ops_rho_x_g4[i][j][t], &ops_rho_x_g4[j][i][(t+t0)%nt]);
		        QLA_C_peq_C_times_C(&corr_rho_g4[t0], &ops_rho_y_g4[i][j][t], &ops_rho_y_g4[j][i][(t+t0)%nt]);
		        QLA_C_peq_C_times_C(&corr_rho_g4[t0], &ops_rho_z_g4[i][j][t], &ops_rho_z_g4[j][i][(t+t0)%nt]);*/
			  }
			}
		}
	  }
	}
  }
  
  qhmc_complex_t corr_ret[nt];
  for (int i = 0; i < nt; i++)
  {
    corr_ret[i].r = QLA_real(corr_scalar[i]);
	corr_ret[i].i = QLA_imag(corr_scalar[i]);
	
  }
  
  /*
  qhmc_complex_t corr_ret[4*nt];
  for (int i = 0; i < nt; i++)
  {
    corr_ret[i].r = QLA_real(corr_pion_gs[i]);
	corr_ret[i].i = QLA_imag(corr_pion_gs[i]);
	corr_ret[nt+i].r = QLA_real(corr_scalar[i]);
	corr_ret[nt+i].i = QLA_imag(corr_scalar[i]);
	corr_ret[2*nt+i].r = QLA_real(corr_rho[i])/3.0;
	corr_ret[2*nt+i].i = QLA_imag(corr_rho[i])/3.0;
	corr_ret[3*nt+i].r = QLA_real(corr_rho_g4[i])/3.0;
	corr_ret[3*nt+i].i = QLA_imag(corr_rho_g4[i])/3.0;
  }
  */
  
  push_complex_array(L, nt, corr_ret);
  //push_complex_array(L, 4*nt, corr_ret);
  
  return 1;
}

static struct luaL_Reg squark_reg[] = {
  { "__gc",          qopqdp_squark_gc },
  { "clone",         qopqdp_squark_clone },
  { "zero",          qopqdp_squark_zero },
  { "point",         qopqdp_squark_point },
  { "random",        qopqdp_squark_random },
  { "randomU1",      qopqdp_squark_randomU1 },
  { "randomZ2",      qopqdp_squark_randomZ2 },
  { "load",          qopqdp_squark_load },
  { "save",          qopqdp_squark_save },
  { "set",           qopqdp_squark_set },
  { "conj",			 qopqdp_squark_conj },
  { "dilute",        qopqdp_squark_dilute }, // ESW addition 3/20/2014, update 5/12/2014
  { "combine",       qopqdp_squark_combine },
  { "getSite",       qopqdp_squark_getSite },
  { "norm2",         qopqdp_squark_norm2 },
  { "norm2_wallsink",qopqdp_squark_norm2_wallsink },
  { "reDot",         qopqdp_squark_redot },
  { "Re_dot",        qopqdp_squark_redot },
  { "dot",           qopqdp_squark_dot },
  { "wall",          qopqdp_squark_wall }, // ESW addition 10/8/2013
  { "wall_gaussian", qopqdp_squark_wall_gaussian }, // ESW addition 2/12/2014
  { "wall_corner",   qopqdp_squark_wall_corner }, // ESW addition 3/3/2014
  { "rephase",       qopqdp_squark_rephase }, // ESW addition 8/23/2013
  { "symshift",      qopqdp_squark_symshift },
  { "epsContract",   qopqdp_squark_epscontr }, // ESW addition 11/15/2013
  { "sum_lattice",   qopqdp_squark_sum_lattice }, // ESW addition 3/3/2014
  { "s4Ferm",        qopqdp_squark_s4_ferm_observables }, // ESW addition 12/18/2013
  { "stoch_conn",    qopqdp_squark_stoch_connected }, // ESW addition 5/13/2014
  { NULL, NULL}
};

squark_t *
qopqdp_squark_create_unset(lua_State *L, int nc, lattice_t *lat)
{
#define NC nc
  if(nc==0) nc = DEFAULTNC;
  if(lat==NULL) lat = qopqdp_get_default_lattice(L);
  squark_t *q = lua_newuserdata(L, sizeof(squark_t));
  q->cv = QDP_create_V_L(lat->qlat);
  q->lat = lat;
  q->qlat = lat->qlat;
  q->nc = nc;
  if(luaL_newmetatable(L, mtname)) {
    lua_pushvalue(L, -1);
    lua_setfield(L, -2, "__index");
    luaL_register(L, NULL, squark_reg);
  }
  lua_setmetatable(L, -2);
  return q;
}

squark_t *
qopqdp_squark_create(lua_State *L, int nc, lattice_t *lat)
{
  squark_t *q = qopqdp_squark_create_unset(L, nc, lat);
  QDP_V_eq_zero(q->cv, QDP_all);
  return q;
}
